#!/usr/bin/env perl
use strict;
use warnings;
use Getopt::Long;
use IO::File;
use JSON::MaybeXS;
use Crypt::LE ':errors';

sub info { print shift, "\n" }
sub bail { &info; exit 255 }

my $VERSION = '0.05';

my ($account_key, $csr, $csr_key, $domains, $challenge_path, $crt, $live, $generate_missing, $generate_only, 
    $handle_with, $handle_as, $handle_params, $complete_with, $complete_params, $verified, $debug, $help);

GetOptions ("key=s" => \$account_key, "csr=s" => \$csr, "csr-key=s" => \$csr_key, "domains=s" => \$domains, "path=s" => \$challenge_path,
            "crt=s" => \$crt, "generate-missing" => \$generate_missing, "generate-only" => \$generate_only, "handle-with=s" => \$handle_with,
	    "handle-as=s" => \$handle_as, "handle-params=s" => \$handle_params, "complete-with=s" => \$complete_with, 
            "complete-params=s" => \$complete_params, "live" => \$live, "debug" => \$debug, "verified" => \$verified, "help" => \$help);

my $usage = <<EOF;

Usage example: le.pl --key account.key --csr domain.csr --csr-key domain.key --crt domain.crt --domains "www.domain.ext,domain.ext" --generate-missing

Note: by default a staging server is used, which does not provide trusted certificates. This is to aviod hitting a rate limits on Let's Encrypt
      live server. To generate an actual certificate, always add --live option.

Available parameters:

 key <file>                       - Your account key file.
 csr <file>                       - Your CSR file.
 csr-key <file>                   - Key for your CSR (only mandatory if CSR is missing and to be generated).
 domains <list>                   - Domains as comma-separated list (only mandatory if CSR is missing).
 path <absolute path>             - Path to local .well-known/acme-challenge/ to drop required challenge files into (optional).
 handle-with <Some::Module>       - Module name to handle challenges with (optional).
 handle-as <http|dns|tls|...>     - Type of challenge to request, by default 'http' (optional).
 handle-params <{json}>           - JSON with parameters to be passed to the challenge-handling module (optional).
 complete-with <Another::Module>  - Module name to handle process completion with (optional).
 complete-params <{json}>         - JSON with parameters to be passed to the completion-handling module (optional).
 generate-missing                 - Generate missing files (key, csr and csr-key).
 generate-only                    - Generate a new key and/or CSR if they are missing and then exit.
 crt <file>                       - Name for the domain certificate file.
 verified                         - Skip challenge verification steps (request/accept/verify). Use only if domains have been already verified recently.
 live                             - Connect to a live server instead of staging.
 debug                            - Print out debug messages.
 help                             - This screen.

[ See Crypt::LE::Challenge::Simple for an example of challenge handling module. ]
[ See Crypt::LE::Complete::Simple for an example of completion handling module. ]

EOF

info "\nCrypt::LE client v$VERSION started.\n";

bail $usage if $help;

unless (($account_key and (-r $account_key or $generate_missing)) and
        ($csr and (-r $csr or ($csr_key and $generate_missing)))) {
    info "Incorrect parameters - need both an account key and CSR either loaded or generated.";
    bail $usage;
}

if (!$generate_missing and ! -r $csr and (!$domains or $domains=~/^\s*$/)) {
    bail "Domain list should be provided to generate a CSR.";
}

unless ($crt) {
    bail "You would probably want to save the domain certificate somewhere - please specify a filename";
}

if ($challenge_path) {
    bail "Path to save challenge files into should be a writable directory" unless (-d $challenge_path and -w _);
}

my ($handler, $complete_handler);
$handle_as = $handle_as ? lc($handle_as) : 'http';

if ($handle_with) {
    eval "use $handle_with;";
    bail "Cannot use the module to handle challenges with." if $@;
    $handler = new $handle_with;
    my $method = 'handle_challenge_' . $handle_as;
    bail "Module to handle challenges does not seem to support the challenge type of '$handle_as'." unless $handler->can($method);
    if ($handle_params) {
	my $j = JSON->new->canonical()->allow_nonref();
        eval {
	    $handle_params = $j->decode($handle_params);
        };
        bail "Could not decode parameters for the challenge-handling module: $@" if $@;
        bail "Please make sure that your 'handle-params' JSON document has {} around it." unless (ref $handle_params eq 'HASH');
    }
} else {
    undef $handle_params;
}

if ($complete_with) {
    eval "use $complete_with;";
    bail "Cannot use the module to complete processing with." if $@;
    $complete_handler = new $complete_with;
    bail "Module to complete processing with does not seem to support the required 'complete' method." unless $complete_handler->can('complete');
    if ($complete_params) {
        my $j = JSON->new->canonical()->allow_nonref();
        eval {
            $complete_params = $j->decode($complete_params);
        };
        bail "Could not decode parameters for the completion module: $@" if $@;
        bail "Please make sure that your 'complete-params' JSON document has {} around it." unless (ref $complete_params eq 'HASH');
    }
} else {
    undef $complete_params;
}

# Begin work

my $le = Crypt::LE->new(debug => $debug, live => $live);

if (-r $account_key) {
    info "Loading an account key from $account_key";
    $le->load_account_key($account_key) == OK or bail "Could not load an account key: " . $le->error_details;
} else {
    info "Generating a new account key";
    $le->generate_account_key == OK or bail "Could not generate an account key: " . $le->error_details;
    info "Saving generated account key into $account_key";
    bail "Failed to save an account key file" if _write($account_key, $le->account_key);
}

if (-r $csr) {
    info "Loading a CSR from $csr";
    $le->load_csr($csr, $domains) == OK or bail "Could not load a CSR: " . $le->error_details;
} else {
    info "Generating a new CSR for domains '$domains'";
    $le->generate_csr($domains) == OK or bail "Could not generate a CSR: " . $le->error_details;
    info "Saving a new CSR into $csr";
    bail "Failed to save a CSR" if _write($csr, $le->csr);
    info "Saving a new CSR key into $csr_key";
    bail "Failed to save a CSR key" if _write($csr_key, $le->csr_key);
}

exit if $generate_only;

bail $le->error_details if $le->register;
$le->accept_tos(); # No need to check explicitly for tos_changed() - Crypt::LE will handle that for you.
unless ($verified) {
    bail $le->error_details if $le->request_challenge();
    bail $le->error_details if $le->accept_challenge($handler ? $handler : \&process_challenge, $handle_params, $handle_as);
    bail $le->error_details if $le->verify_challenge;
}
info "Requesting domain certificate";
bail $le->error_details if $le->request_certificate();
info "Requesting issuer's certificate";
if ($le->request_issuer_certificate()) {
    info "Could not download an issuer's certificate, try to download manually from " . $le->issuer_url;
    info "Will be saving the domain certificate alone, not the full chain (Qualys SSL test score will be capped to 'B' at best)";
    bail "Failed to save the domain certificate file" if _write($crt, $le->certificate);
} else {
    info "Saving the full certificate chain (With the proper server configuration and HSTS you can get an 'A+' on Qualys SSL test)";
    bail "Failed to save the domain certificate file" if _write($crt, $le->certificate . "\n" . $le->issuer);
}
if ($complete_handler) {
    my $data = {
        # Note, certificate here is just a domain certificate, issuer is passed separately - so handler 
        # could merge those or use them separately as well.
        certificate => $le->certificate, certificate_file => $crt, key_file => $csr_key, issuer => $le->issuer,
    };
    my $rv;
    eval {
	$rv = $complete_handler->complete($data, $complete_params);
    };
    if ($@ or !$rv) {
        bail "Completion handler " . ($@ ? "thrown an error: $@" : "did not return a true value");
    }
}

info "The job is done, enjoy your certificate! For feedback and bug reports contact me at [ https://do-know.com ]";

sub _write {
    my ($file, $content) = @_;
    return 1 unless ($file and $content);
    my $fh = IO::File->new($file, 'w');
    return 1 unless defined $fh;
    $fh->binmode;
    print $fh $content;
    $fh->close;
    return 0;
}

sub process_challenge {
    my $challenge = shift;
    my $text = "$challenge->{token}.$challenge->{fingerprint}";
    if ($challenge_path) {
        my $file = "$challenge_path/$challenge->{token}";
	if (_write($file, $text)) {
	   info "Failed to save a challenge file '$file' for domain '$challenge->{domain}'";
           return 0;
	} else {
           info "Successfully saved a challenge file '$file' for domain '$challenge->{domain}'";
           return 1;
        }
    }
    print <<EOF;
Challenge for $challenge->{domain} requires:
A file '$challenge->{token}' in '/.well-known/acme-challenge/' with the text: $text
When done, press <Enter>
EOF
 	<STDIN>;
	return 1;
};
